/*-
 * Copyright (c) 1999,2000 Joao Cabral (jcnc@dhis.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      DHIS(c)  Dynamic Host Information System Release 4.0
 */

#include<sys/types.h>
#include<sys/socket.h>
#include<netdb.h>
#include<netinet/in.h>
#include<arpa/inet.h>
#include<time.h>
#include<ctype.h>
#include<stdio.h>
#include<stdlib.h>
#include<unistd.h>
#include<string.h>
#include<signal.h>
#include<syslog.h>
#include<sys/time.h>

#ifdef	QRC
#include<gmp.h>
#endif

/* Define statements */

#define	DHIS_VERSION	40		/* DHIS Version */

#define	MAX_HOSTNAME	64
#define	MAX_PASS	16		/* Characters in password */

/* Message opcodes */
#define	ECHO_REQ	0x00000411
#define	ECHO_ACK	0x00000412
#define	AUTH_REQ	0x00000421
#define	AUTH_DENY	0x00000422
#define	AUTH_ACK	0x00000423
#define	AUTH_SX		0x00000424
#define	AUTH_SY		0x00000425
#define	AUTH_RE		0x00000426

#define	CTL4_ISADDR	0x00000431
#define	CTL4_MXADDR	0x00000432
#define	CTL4_MXEND	0x00000433
#define	CTL4_MXSTART	0x00000434

#define	CHECK_REQ	0x00000441
#define	CHECK_ACK	0x00000442

#define	OFFLINE_REQ	0x00000451


#define	DHID_CONF	"/etc/dhid.conf"
#define	DHID_PID	"/etc/dhid.pid"

#define	DEF_ISPORT	58800
#define	FAIL_ALLOW	3

#define	APASS		1
#define	AQRC		2

/* Configuration section */

struct conf_t {
		int id;					/* HostID */
		unsigned char pass[MAX_PASS];		/* Password */
		unsigned char host[MAX_HOSTNAME]; 	/* FQDN */
		int atype;
		int isa;				/* IS Address */
		int isp;				/* IS Port */
		int mxa;				/* MX Address */
		int sid;
#ifdef	QRC	
		mpz_t authp;
		mpz_t authq;
#endif
	      };

void update_server(int,int,unsigned char *);
void read_conf(unsigned char *);
unsigned char *line_entry(int,unsigned char *);

/* network section */

#define	MAX_MSG	256

typedef struct {
		int opcode;
		int serial;
		int version;
		int rport;
		} header_t;

typedef struct { header_t hdr;
		unsigned char buff[MAX_MSG-sizeof(header_t)];
	       } msg_t;

typedef struct { header_t hdr; } echo_req_t;
typedef struct { header_t hdr; int oserial; } echo_ack_t;
typedef struct { header_t hdr; int id; unsigned char pass[16]; } auth_req_t;
typedef struct { header_t hdr; } auth_deny_t;
typedef struct { header_t hdr; int sid; } auth_ack_t;
typedef struct { header_t hdr; } auth_re_t;
typedef struct { header_t hdr; int id; unsigned char x[200]; } auth_sendx_t;
typedef struct { header_t hdr; unsigned char y[200]; } auth_sendy_t;
typedef struct { header_t hdr; int isaddr; } ctl4_isaddr_t;
typedef struct { header_t hdr; int mxaddr; } ctl4_mxaddr_t;
typedef struct { header_t hdr; } ctl4_mxend_t;
typedef struct { header_t hdr; int next_check; } check_req_t;
typedef struct { header_t hdr; int sid; } check_ack_t;
typedef struct { header_t hdr; int sid; } offline_req_t;

int msg_size_by_opcode(int);
void swap_int(int *);
void swap_msg(int *,int);
int little_endian(void);
void convert_message(msg_t *,int);
int get_serial(void);
int net_init(int);
int net_close(void);
int net_check_message(void);
int net_read_message(msg_t *,int *);
int net_write_message(msg_t *,int,int);

#ifdef	QRC
void qrc_random(mpz_t,int);
void qrc_genkey(mpz_t);
void qrc_genx(mpz_t,mpz_t);
void qrc_geny(mpz_t,mpz_t,mpz_t);
void qrc_sqrty(mpz_t,mpz_t,mpz_t);
void qrc_crt(mpz_t,mpz_t,mpz_t,mpz_t,mpz_t);
void qrc_fill_str(mpz_t,unsigned char *,int);
#endif

