/*-
 * Copyright (c) 1999,2000 Joao Cabral (jcnc@dhis.org)
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *      DHIS(c)  Dynamic Host Information System Release 4.0
 */

#include "dhid.h"

unsigned char conffile[256];
int relterm=0;
int background=1;
int timeout=-1;
int defconf=0;


extern struct conf_t conf;

int rport=DEF_ISPORT;

#define	POLL_STAGE	1
#define	AUTH_STAGE	2
#define	ON_STAGE	3
int stage=POLL_STAGE;

void sig_usr2() {
	
	if(stage==ON_STAGE) {
		offline_req_t msg;
		msg.hdr.opcode=OFFLINE_REQ;
		msg.sid=conf.sid;
		net_write_message((msg_t *)&msg,conf.isa,conf.isp);
	}
	conf.sid=0;
	stage=POLL_STAGE;
	syslog(LOG_INFO,"-> Offline (USR2)");
	timeout=60;
}

void sig_term() {
	
	
	if(stage==ON_STAGE) {
		offline_req_t msg;
		msg.hdr.opcode=OFFLINE_REQ;
		msg.sid=conf.sid;
		net_write_message((msg_t *)&msg,conf.isa,conf.isp);
		net_close();
	}
	syslog(LOG_INFO,"daemon exiting");
	if(!defconf)
	unlink(DHID_PID);
	exit(0);
}

void usage(char *argv[]) {
	
	printf("Syntax: %s [-i] [-r] [-p port] [-f conf_file]\n",argv[0]);
	exit(0);
} 
	
void check_message(void) {

	msg_t msg;
	int from;

	if(!net_check_message()) return;
	if(!net_read_message(&msg,&from)) return;
	
	if(msg.hdr.version<DHIS_VERSION) return;

	if(msg.hdr.opcode==ECHO_REQ) {
		echo_ack_t m;
		m.hdr.opcode=ECHO_ACK;
		m.oserial=msg.hdr.serial;
		net_write_message((msg_t *)&m,from,msg.hdr.rport);
		return;
	}

	if(msg.hdr.opcode==AUTH_DENY) {
		syslog(LOG_ERR,"Authentication failed, retrying ...");
		stage=POLL_STAGE;
		timeout=30;
		return;
	}

	if(msg.hdr.opcode==AUTH_ACK) {
		auth_ack_t *p2;
		p2=(auth_ack_t *)&msg;
		timeout=60*FAIL_ALLOW;
		stage=ON_STAGE;	
		syslog(LOG_INFO,"-> Online");
		conf.sid=p2->sid;
		return;
	}
	if(msg.hdr.opcode==AUTH_RE) {
		if(conf.isa==from) {
		conf.sid=0;
		stage=POLL_STAGE;
		syslog(LOG_INFO,"Authentication requested");
		timeout=0;
		}
		return;
	}
		

	if(msg.hdr.opcode==CTL4_ISADDR) {
		ctl4_isaddr_t *p; p=(ctl4_isaddr_t *)&msg;
		update_server(CTL4_ISADDR,p->isaddr,conffile);
		return;
	}

	if(msg.hdr.opcode==CTL4_MXADDR) {
		ctl4_mxaddr_t *p; p=(ctl4_mxaddr_t *)&msg;
		update_server(CTL4_MXADDR,p->mxaddr,conffile);
		return;
	}

	if(msg.hdr.opcode==CTL4_MXEND) {
		if(relterm) {
		syslog(LOG_INFO,"MXEND received. Exiting ...");
		net_close();
		if(!defconf)
		unlink(DHID_PID);
		exit(0);
		}
		return;
	}

	if(msg.hdr.opcode==CHECK_REQ) {
		check_req_t *p;
		check_ack_t m;
		if(stage!=ON_STAGE) { stage=POLL_STAGE; return; }
		p=(check_req_t *)&msg;
		m.hdr.opcode=CHECK_ACK;
		m.sid=conf.sid;
		net_write_message((msg_t *)&m,from,msg.hdr.rport);
		timeout=FAIL_ALLOW*p->next_check;
		return;
	}

	if(msg.hdr.opcode==ECHO_ACK) {
		auth_req_t m;
		stage=AUTH_STAGE;
		if(conf.atype==APASS) strcpy(m.pass,conf.pass);
		else memset(m.pass,0,16);
		m.hdr.opcode=AUTH_REQ;
		m.id=conf.id;
		net_write_message((msg_t *)&m,from,msg.hdr.rport);
		timeout=30;
		return;
	}

#ifdef	QRC
	if(msg.hdr.opcode==AUTH_SY) {
		mpz_t x1,x2,y,x,n;
		auth_sendx_t m;
		auth_sendy_t *mp;
		unsigned char buff[1024];
		
		if(conf.atype!=AQRC) return;

		mpz_init(x1);
		mpz_init(x2);
		mpz_init(y);
		mpz_init(x);
		mpz_init(n);

		mp=(auth_sendy_t *)&msg;
		memcpy(buff,mp->y,200);
		buff[200]='\0';
		mpz_set_str(y,buff,10);

		qrc_sqrty(x1,y,conf.authp);
		qrc_sqrty(x2,y,conf.authq);

		qrc_crt(y,x1,x2,conf.authp,conf.authq);

		mpz_clear(x1);
		mpz_clear(x2);

		mpz_mul(n,conf.authp,conf.authq);
		
		mpz_mod(x,y,n);

		mpz_clear(y);
		mpz_clear(n);
		
		m.hdr.opcode=AUTH_SX;
		m.id=conf.id;
		qrc_fill_str(x,m.x,200);
		mpz_clear(x);
		net_write_message((msg_t *)&m,from,msg.hdr.rport);
		return;
	}
#endif

	
	return;

}


main(int argc,char *argv[]) {

	FILE *fp;
	int c;
	extern char *optarg;

	strcpy(conffile,DHID_CONF);

	while((c=getopt(argc,argv,"f:ihrp:"))!=EOF)
	switch(c) {
	case('i'): background=0; break;
	case('r'): relterm=1; break;
	case('p'): rport=atoi(optarg); break;
	case('f'): defconf=1;strncpy(conffile,optarg,sizeof(conffile)); break;
	case('h'): usage(argv);
	}

	
	if(getuid()) {
		syslog(LOG_ERR,"%s must be executed by root",argv[0]);
		exit(1);
	}


	close(0);
	close(1);
	close(2);

	if(background)
	if(fork()) exit(0);	/* Run in background */
	


	syslog(LOG_INFO,"daemon started");

	read_conf(conffile);
	
	if(net_init(rport)) {
		syslog(LOG_ERR,"failed to open incoming udp socket on port %d",
			rport);
		exit(255);
	}
	
	signal(SIGUSR1,SIG_IGN);
	signal(SIGINT,SIG_IGN);
	signal(SIGUSR2,sig_usr2);
	signal(SIGTERM,sig_term);

	if(!defconf) {
	unlink(DHID_PID);
	fp=fopen(DHID_PID,"w");
	if(fp!=NULL) {
		fprintf(fp,"%d",getpid());
		fclose(fp);
	}
	}


	
	for(;;) {
		check_message();
		if(stage==POLL_STAGE && timeout<=0) {
			echo_req_t m; m.hdr.opcode=ECHO_REQ;
			net_write_message((msg_t *)&m,conf.isa,conf.isp);
			timeout=30;
		}
		if(timeout>0) timeout--;
		if(timeout==0) { 
			if(stage==ON_STAGE) 
			syslog(LOG_INFO,"-> Offline (timeout)");
			stage=POLL_STAGE;
		}
		sleep(1);
	}
}
